# description from
# https://github.com/neago/lecroy-reader/blob/49c42a85c449013c1c48d154ae70192f172e32ba/lecroyreader/lecroy.py#L4
trc_description = (
    ("descriptor_name", "16s"),
    ("template_name", "16s"),
    ("comm_type", "h"),
    ("comm_order", "h"),
    ("wave_descriptor", "i"),
    ("user_text", "i"),
    ("res_desc1", "i"),
    ("trig_time_array", "i"),
    ("ris_time_array", "i"),
    ("res_array1", "i"),
    ("wave_array1", "i"),
    ("wave_array2", "i"),
    ("res_array2", "i"),
    ("res_array3", "i"),
    ("instrument_name", "16s"),
    ("instrument_number", "i"),
    ("trace_label", "16s"),
    ("reserved1", "h"),
    ("reserved2", "h"),
    ("wave_array_count", "i"),
    ("points_per_screen", "i"),
    ("first_valid_point", "i"),
    ("last_valid_point", "i"),
    ("first_point", "i"),
    ("sparsing_factor", "i"),
    ("segment_index", "i"),
    ("subarray_count", "i"),
    ("sweeps_per_acq", "i"),
    ("points_per_pair", "h"),
    ("pair_offset", "h"),
    ("vertical_gain", "f"),
    ("vertical_offset", "f"),
    ("max_value", "f"),
    ("min_value", "f"),
    ("nominal_bits", "h"),
    ("nom_subarray_count", "h"),
    ("horiz_interval", "f"),
    ("horiz_offset", "d"),
    ("pixel_offset", "d"),
    ("vert_unit", "48s"),
    ("horiz_unit", "48s"),
    ("horiz_uncertainty", "f"),
    ("trigger_time", "dbbbbhh"),
    ("acq_duration", "f"),
    ("record_type", "h"),
    ("processing_done", "h"),
    ("reserved5", "h"),
    ("ris_sweeps", "h"),
    ("time_base", "h"),
    ("vert_coupling", "h"),
    ("probe_att", "f"),
    ("fixed_vert_gain", "h"),
    ("bandwidth_limit", "h"),
    ("vertical_vernier", "f"),
    ("acq_vert_offset", "f"),
    ("wave_source", "h"),
)

_trc_description_fields = {name for name, _ in trc_description}


class Header:
    def __init__(self, header: dict):
        for name in _trc_description_fields:
            setattr(self, f"_{name}", header[name])

    def __iter__(self):
        for name in _trc_description_fields:
            yield name, getattr(self, f"_{name}")

    def __getitem__(self, item):
        if item not in _trc_description_fields:
            raise KeyError(
                f"Invalid header field: {item}. Valid fields are: {_trc_description_fields}"
            )
        return getattr(self, f"_{item}")

    def __str__(self):
        return f"""Instrument name: {self["instrument_name"]}
Trigger time: {self["trigger_time"]}
Vertical coupling: {self["vert_coupling"]}
Time base: {self["time_base"]}
Horizontal interval: {self["horiz_interval"]}
Horizontal offset: {self["horiz_offset"]}
Fixed vertical gain: {self["fixed_vert_gain"]}
Vertical gain: {self["vertical_gain"]}
Vertical offset: {self["vertical_offset"]}
Wave array count: {self["wave_array_count"]}
Subarray count: {self["subarray_count"]}"""

    @property
    def sequence(self):
        return self["subarray_count"] > 1


# add header fields as properties
for _name in _trc_description_fields:
    setattr(
        Header,
        _name,
        property(lambda self, name=_name: getattr(self, f"_{name}")),
    )
del _name



from os import PathLike

import numpy

from .file import read
from .header import Header


class Trace:
    def __init__(
        self, filename_or_bytes: str | PathLike[str] | bytes, header_only: bool = False
    ):

        self._filename = (
            filename_or_bytes if not isinstance(filename_or_bytes, bytes) else ""
        )

        header, self._trigger_times, self._voltage = read(
            filename_or_bytes, header_only
        )
        self._header = Header(header)

        # store values in voltage units
        self._voltage = self._voltage * self._header["vertical_gain"]
        self._voltage = self._voltage + self._header["vertical_offset"]

        # compute time values
        self._time = (
            numpy.arange(self._voltage.shape[-1]) * self._header["horiz_interval"]
            + self._header["horiz_offset"]
        )

    def __len__(self):
        if not self._header.sequence:
            return 1
        return len(self._voltage)

    def __iter__(self):
        if len(self) == 1:
            yield self._time, self._voltage
        else:
            for single in self._voltage:
                yield self._time, single

    @property
    def header(self):
        return self._header

    @property
    def header_only(self):
        return self._voltage.size == 0

    @property
    def voltage(self):
        return self._voltage

    @property
    def trigger_times(self):
        return self._trigger_times

    @property
    def time(self):
        return self._time

    # Alternative names
    @property
    def x(self):
        return self._time

    @property
    def y(self):
        return self._voltage



import mmap
from os import PathLike
import struct
import numpy
from io import BytesIO
from datetime import datetime

from .header import trc_description


def read(
    filename_or_bytes: str | PathLike[str] | bytes, header_only: bool = False
) -> tuple[dict[str, str | int | float], numpy.ndarray, numpy.ndarray]:
    with open(filename_or_bytes, "r+b") if not isinstance(
        filename_or_bytes, bytes
    ) else BytesIO(filename_or_bytes) as f:

        wavedesc_bytes = b"WAVEDESC"
        # find "WAVEDESC" and skip those bytes
        if isinstance(filename_or_bytes, bytes):
            f.read(filename_or_bytes.find(wavedesc_bytes))
        else:
            # https://docs.python.org/3/library/mmap.html
            mm = mmap.mmap(f.fileno(), 0)
            f.read(mm.find(wavedesc_bytes))

        header = {
            name: f.read(struct.Struct(fmt).size) for (name, fmt) in trc_description
        }

        endianness = ">" if struct.unpack("h", header["comm_order"])[0] == 0 else "<"

        header = {
            name: struct.unpack(f"{endianness}{fmt}", header[name])
            for (name, fmt) in trc_description
        }

        for name in header:
            if name != "trigger_time":
                header[name] = header[name][0]
            else:
                trigger_time = reversed(header["trigger_time"][:-1])
                # convert seconds to integer (lose milliseconds)
                trigger_time = [int(t) for t in trigger_time]
                header[name] = datetime(*trigger_time).isoformat()

        # format strings
        for name in header:
            if isinstance(header[name], bytes):
                header[name] = header[name].decode("ascii").strip("\x00")

        # format some attributes in a more human-readable way
        # https://github.com/neago/lecroy-reader/blob/49c42a85c449013c1c48d154ae70192f172e32ba/lecroyreader/lecroy.py
        record_types = [
            "single sweep",
            "interleaved",
            "histogram",
            "graph",
            "filter coefficient",
            "complex",
            "extrema",
            "sequence obsolete",
            "centered RIS",
            "peak detect",
        ]
        header["record_type"] = record_types[header["record_type"]]

        processing_types = [
            "no processing",
            "fir filter",
            "interpolated",
            "sparsed",
            "autoscaled",
            "no result",
            "rolling",
            "cumulative",
        ]
        header["processing_done"] = processing_types[header["processing_done"]]

        vertical_couplings = [
            "DC 50 Ohm",
            "ground",
            "DC 1 MOhm",
            "ground",
            "AC 1 MOhm",
        ]
        header["vert_coupling"] = vertical_couplings[header["vert_coupling"]]
        time_base = header["time_base"]
        if time_base == 100:
            header["time_base"] = "external"
        else:
            value = [1, 2, 5, 10, 20, 50, 100, 200, 500][time_base % 9]
            prefix = ["p", "n", "μ", "m", "", "k"][time_base // 9]
            header["time_base"] = f"{value} {prefix}s / div"

        fixed_vert_gain = header["fixed_vert_gain"]
        value = [1, 2, 5, 10, 20, 50, 100, 200, 500][fixed_vert_gain % 9]
        prefix = ["μ", "m", "", "k"][fixed_vert_gain // 9]
        header["fixed_vert_gain"] = f"{value} {prefix}V / div"

        # if header only return empty arrays
        values_type = numpy.int8 if header["comm_type"] == 0 else numpy.int16
        if not header_only:
            if header["user_text"] != 0:
                # skip user text
                f.read(header["user_text"])

            trigger_times = numpy.frombuffer(
                f.read(int(header["trig_time_array"])), dtype=numpy.float64
            )

            values = numpy.frombuffer(
                f.read(int(header["wave_array_count"])), dtype=values_type
            )
        else:
            trigger_times = numpy.array([], dtype=numpy.float64)
            values = numpy.array([], dtype=values_type)

        trigger_times = trigger_times.reshape((2, -1), order="F")
        if (subarray_count := header["subarray_count"]) > 1:
            values = values.reshape((subarray_count, -1), order="C")

        return header, trigger_times, values



import pytest

from numpy.testing import assert_array_equal
from pathlib import Path
import numpy as np

import lecroyscope

from .test_header import header_reference_dict

files_path = Path(__file__).parent / "files"


def test_read_header():
    (
        header_header_only,
        trigger_times_header_only,
        values_header_only,
    ) = lecroyscope.reading.read(files_path / "header.trc", header_only=True)

    assert header_header_only == header_reference_dict

    header, trigger_times, values = lecroyscope.reading.read(
        files_path / "header.trc", header_only=False
    )

    assert header == header_header_only
    assert_array_equal(trigger_times, trigger_times_header_only)
    assert_array_equal(values, values_header_only)

    assert trigger_times.shape[0] == 2

    assert values.dtype in (np.int8, np.int16)


def test_read_header_from_bytes():
    filename = files_path / "header.trc"
    (
        header_from_file,
        trigger_times_from_file,
        values_from_file,
    ) = lecroyscope.reading.read(files_path / "header.trc", header_only=True)

    (
        header_from_bytes,
        trigger_times_from_bytes,
        values_from_bytes,
    ) = lecroyscope.reading.read(filename.read_bytes(), header_only=True)

    assert_array_equal(values_from_file, values_from_bytes)
    assert_array_equal(trigger_times_from_file, trigger_times_from_bytes)
    assert header_from_file == header_from_bytes



import pytest

import lecroyscope

header_reference_dict = {
    "descriptor_name": "WAVEDESC",
    "template_name": "LECROY_2_3",
    "comm_type": 1,
    "comm_order": 1,
    "wave_descriptor": 346,
    "user_text": 0,
    "res_desc1": 0,
    "trig_time_array": 3200,
    "ris_time_array": 0,
    "res_array1": 0,
    "wave_array1": 800800,
    "wave_array2": 0,
    "res_array2": 0,
    "res_array3": 0,
    "instrument_name": "LECROYWR64Xi-A",
    "instrument_number": 50699,
    "trace_label": "",
    "reserved1": 7184,
    "reserved2": 6,
    "wave_array_count": 400400,
    "points_per_screen": 400000,
    "first_valid_point": 0,
    "last_valid_point": 400399,
    "first_point": 0,
    "sparsing_factor": 1,
    "segment_index": 0,
    "subarray_count": 200,
    "sweeps_per_acq": 1,
    "points_per_pair": 0,
    "pair_offset": 0,
    "vertical_gain": 0.00012499500007834285,
    "vertical_offset": -0.949999988079071,
    "max_value": 31745.0,
    "min_value": -32001.0,
    "nominal_bits": 8,
    "nom_subarray_count": 200,
    "horiz_interval": 9.999999717180685e-10,
    "horiz_offset": -2.2824463729809135e-07,
    "pixel_offset": -2.2800000000000008e-07,
    "vert_unit": "V",
    "horiz_unit": "S",
    "horiz_uncertainty": 9.999999960041972e-13,
    "trigger_time": "2022-10-13T16:29:38",
    "acq_duration": 0.0,
    "record_type": "single sweep",
    "processing_done": "no processing",
    "reserved5": 0,
    "ris_sweeps": 1,
    "time_base": "200 ns / div",
    "vert_coupling": "DC 50 Ohm",
    "probe_att": 1.0,
    "fixed_vert_gain": "1 V / div",
    "bandwidth_limit": 0,
    "vertical_vernier": 1.0,
    "acq_vert_offset": -0.949999988079071,
    "wave_source": 1,
}


def test_header_dict():
    header = lecroyscope.reading.Header(header_reference_dict)
    header_as_dict = dict(header)
    assert header_as_dict == header_reference_dict


def test_header_getitem():
    header = lecroyscope.reading.Header(header_reference_dict)
    assert header["horiz_interval"] == header_reference_dict["horiz_interval"]

    with pytest.raises(KeyError):
        header["does_not_exist"]


def test_header_string():
    header = lecroyscope.reading.Header(header_reference_dict)
    assert (
        str(header)
        == """Instrument name: LECROYWR64Xi-A
Trigger time: 2022-10-13T16:29:38
Vertical coupling: DC 50 Ohm
Time base: 200 ns / div
Horizontal interval: 9.999999717180685e-10
Horizontal offset: -2.2824463729809135e-07
Fixed vertical gain: 1 V / div
Vertical gain: 0.00012499500007834285
Vertical offset: -0.949999988079071
Wave array count: 400400
Subarray count: 200"""
    )



import pytest

from numpy.testing import assert_array_equal
import numpy as np
from pathlib import Path

import lecroyscope

from .test_header import header_reference_dict

files_path = Path(__file__).parent / "files"


def test_trace_header():
    filename = files_path / "header.trc"
    trace = lecroyscope.Trace(filename, header_only=True)
    assert dict(trace.header) == header_reference_dict


def test_trace_iter():
    filename = files_path / "header.trc"
    trace = lecroyscope.Trace(filename)

    # this header file is from a sequence of size 200
    assert len(trace) == 200

    data = np.array([(time, single) for time, single in trace])
    assert len(data) == len(trace)
    times = data[:, 0]
    voltage = data[:, 1]
    assert len(times) == len(voltage) == len(trace)

    # check all times are equal
    for time in times:
        assert_array_equal(time, times[0])
